/*
 *      Copyright 2006-2006 Richard Gorton - All rights reserved
 *      Exclusively licensed to Gorton Studios Inc.
 *
 * utils.c - utility functions for Web Site Consistency Checker (WSLC)
 *
 * Contents:
 *      Site error handling functions
 *      Internal (coding) consistency error handling functions
 *      Hash functions
 *      Parser helper functions
 */
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>

#include "wslc.h"

extern FILREC *HTable[NCHAINS];

typedef struct TYPE_STATS {
        enum WSLC_TYPE  type;
        int             count;
        const char      *name;
        } TYPE_STATS;

static TYPE_STATS TypeStats[WSLC_LAST] = {
        { WSLC_UNKNOWN, 0, "Unknown file type   " },
        { WSLC_ERROR,   0, "Error analyzing file" },
        { WSLC_SDP,     0, "Socket, Device, Pipe" },
        { WSLC_SETX,    0, "setuid or setgid    " },
        { WSLC_LINK,    0, "symbolic link       " },
        { WSLC_DIR,     0, "directory           " },
        { WSLC_SCRIPT,  0, "shell script        " },
        { WSLC_ELF,     0, "ELF executable      " },
        { WSLC_GRAPHICS,0, "Graphics image      " },
        { WSLC_HTML,    0, "[probably] html     " },
        { WSLC_DOCUMENT,0, "MS Office or pdf doc" },
        { WSLC_GZ_COMP, 0, "gzipped/compressed  " },
        { WSLC_LAST,    0, "OOPS! This should never happen!" }
        };

/*
 * ReportSiteError - report a semantic problem with the web tree to be analyzed.
 */
void ReportSiteError (char *msg)
{
    fprintf (stderr, "%s\n", msg);
}

void Error (char *Emsg)
{
    fprintf (stderr, "%s\n", Emsg);
}

void ErrorD (char *fmts, int var)
{
    char    ebuf[256];

    snprintf (ebuf, sizeof(ebuf)-1, fmts, var);
    ebuf[255] = 0x0;
    Error (ebuf);
}

void ErrorS (char *fmts, char *var)
{
    char    ebuf[256];

    snprintf (ebuf, sizeof(ebuf)-1, fmts, var);
    ebuf[255] = 0x0;
    Error (ebuf);
}

/*
 * Straight ripoff of sdbm hash function, except for the multiplier
 */
long sdbm_hash (char *str, int len)
{
    register unsigned long n = 0;

#define HASHC   n = *str++ + 0x10833 * n
 
    if (len > 0) {
        register int loop = (len + 8 - 1) >> 3;
 
        switch(len & (8 - 1)) {
        case 0: do {
                HASHC;  case 7: HASHC;
        case 6: HASHC;  case 5: HASHC;
        case 4: HASHC;  case 3: HASHC;
        case 2: HASHC;  case 1: HASHC;
                } while (--loop);
        }
 
    }

    return n;
}

char *GetNextScanPoint (char *bufstart, int *lineno)
{
    char *c;

    c = bufstart;
    while (*c && (*c != '<')) {
        if (*c == '\n')
	    *lineno += 1;
	c++;
    }

    return c;
}

/*
 * If there is a consistency error, report the error, and return a null
 */
FILREC *FindRecByName (char *fname)
{
    const char *func = "FindRecByName";
    char        errbuf[256];
    int	flen;
    int	fhash;
    int	t_idx;
    FILREC	*walker;

    flen = strlen(fname);
    fhash = sdbm_hash (fname, flen);
    t_idx = fhash & (NCHAINS - 1);
    walker = HTable[t_idx];
    while (walker) {
	if ((walker->hash == fhash)
	    && (strncmp (walker->filepath, fname, flen) == 0)
	   ) {
	    return walker;
	}
	walker = walker->next;
    }
    sprintf (errbuf, "%s (%s)", func, fname);
    InternalError (errbuf, 1, IERR_HASHTAB);
    return NULL;
}

enum WSLC_TYPE GetFileType (FILREC *frec)
{
    return frec->type;
}

// Guess what file type we've got here.  Some we can tell by the magic number
// but others we need to guess based solely upon file extension (if any)
enum WSLC_TYPE GuessFileType (char *Fname, int FnameLen)
{
    enum WSLC_TYPE	rval;
    struct stat	st_b;
    FILE	*this_f;
    int	i;
    char	hdrb[8];

    this_f = NULL;
    rval = WSLC_UNKNOWN; 
    if (stat (Fname, &st_b)) {
	// printf ("File %s is Error\n", Fname);
	rval = WSLC_ERROR;
	goto done;
    }

    if (S_ISSOCK (st_b.st_mode) || S_ISCHR (st_b.st_mode)
	|| S_ISBLK (st_b.st_mode) || S_ISFIFO (st_b.st_mode)) {
	// printf ("File %s is SDP\n", Fname);
	rval = WSLC_SDP;
	goto done;
    }

    if (S_ISLNK (st_b.st_mode)) {
	// printf ("File %s is symlink\n", Fname);
	rval = WSLC_LINK;
	goto done;
    }

    if (st_b.st_mode & (S_ISUID | S_ISGID)) {
	// printf ("File %s is setx\n", Fname);
	rval = WSLC_SETX;
	goto done;
    }
    if (st_b.st_mode & S_IFDIR) {
	// printf ("File %s is dir\n", Fname);
	rval = WSLC_DIR;
	goto done;
    }
	// We can't really tell what it is, might be a null file.
    if (st_b.st_size < 8) {
	// printf ("File %s is tiny: error\n", Fname);
	rval = WSLC_ERROR;
	goto done;
    }

    this_f = fopen (Fname, "r");
    if (!this_f) {
	rval = WSLC_ERROR;
	goto done;
    }
    if (fread ((void *)&hdrb, 1, sizeof(hdrb), this_f) != sizeof (hdrb)) {
	rval = WSLC_ERROR;
	goto done;
    }

    if (memcmp (hdrb, "#!", 2) == 0) {
	// printf ("File %s is script\n", Fname);
	rval = WSLC_SCRIPT;
	goto done;
    }
	// 0x7f E L F
    if ((hdrb[0] == 0x7f) && (hdrb[1] == 'E')
	&& (hdrb[2] == 'L') && (hdrb[3] == 'F')) {
	// printf ("File %s is ELF\n", Fname);
	rval = WSLC_ELF;
	goto done;
    }

	// JPEG, JPG
    if ((hdrb[0] == (char) 0xff) && (hdrb[1] == (char) 0xd8)
	&& (hdrb[2] == (char) 0xff)
#if 0
	// byte 4 seems to be variable for valid variants - ignore it
		// JPEG				JPG
	&& ((hdrb[3] == (char) 0xe0) || (hdrb[3] == (char) 0xee))
#endif
	) {
	// printf ("File %s is JPEG\n", Fname);
	rval = WSLC_GRAPHICS;
	goto done;
    }

	// PNG
    if ((hdrb[0] == (char) 0x89) && (hdrb[1] == 'P')
	&& (hdrb[2] == 'N') && (hdrb[3] == 'G')) {
	// printf ("File %s is PNG\n", Fname);
	rval = WSLC_GRAPHICS;
	goto done;
    }

	// GIF
    if ((hdrb[0] == 'G') && (hdrb[1] == 'I')
	&& (hdrb[2] == 'F') && (hdrb[3] == '8')) {
	// printf ("File %s is GIF\n", Fname);
	rval = WSLC_GRAPHICS;
	goto done;
    }

	// TIFF: big-endian
    if ((hdrb[0] == 'M') && (hdrb[1] == 'M')
	&& (hdrb[2] == 0x00) && (hdrb[3] == 0x2a)) {
	// printf ("File %s is GIF\n", Fname);
	rval = WSLC_GRAPHICS;
	goto done;
    }

	// TIFF: little-endian
    if ((hdrb[0] == 'I') && (hdrb[1] == 'I')
	&& (hdrb[2] == 0x2a) && (hdrb[3] == 0x00)) {
	// printf ("File %s is GIF\n", Fname);
	rval = WSLC_GRAPHICS;
	goto done;
    }

	// PDF
    if ((hdrb[0] == '%') && (hdrb[1] == 'P')
	&& (hdrb[2] == 'D') && (hdrb[3] == 'F')) {
	// printf ("File %s is GIF\n", Fname);
	rval = WSLC_DOCUMENT;
	goto done;
    }

	// MS Office (doesn't include older MS Word formats)
    if ((hdrb[0] == (char) 0xd0) && (hdrb[1] == (char) 0xcf)
	&& (hdrb[2] == (char) 0x11) && (hdrb[3] == (char) 0xe0)) {
	// printf ("File %s is GIF\n", Fname);
	rval = WSLC_DOCUMENT;
	goto done;
    }

        // gzipped or compressed file
    if ((hdrb[0] == (char) 0x1f) &&
        ((hdrb[1] == (char) 0x8b) || (hdrb[1] == (char) 0x9d))) {
	// printf ("File %s is gzipped/compressed\n", Fname);
	rval = WSLC_GZ_COMP;
	goto done;
    }

	// Back up to last "."
    for (i = FnameLen - 1; i >= 0; i--) {
	if (Fname[i] == '.')
	    break;
    }

    if (i <= 0)		// No file extension
	goto done;

	// Move forward one for start of extension compares
    i++;

	// htm, html
    if (strncasecmp ((const char *)&Fname[i], "htm", 3) == 0) {
	rval = WSLC_HTML;
	goto done;
    }

done:
    TypeStats[rval - 1].count++;
    if (this_f)
	fclose (this_f);

    return rval;
}

void ReportTypeStats()
{
    int i;

    printf ("Statistics about the on-disk web site\n");
    printf ("File type\t\tnumber\n"
            "---------\t\t------\n");
    for (i = 0; i < WSLC_LAST - 1; i++) {
        if (TypeStats[i].count > 0) {
            printf ("%s\t%d\n", TypeStats[i].name, TypeStats[i].count);
        }
    }
}

/*
 * Internal consistency error handler
 *      func - function name
 *      fn_e - error point within the function (1..n)
 *      code - type of error
 */
void InternalError (const char *func, int fn_e, enum IERR_CODES code)
{
    int i;
    char    *code_s;

    for (i = 0; i < IERR_LAST; i++) {
        code_s = (char *)IntErrString[i].string;
        if (IntErrString[i].code == code) {
            break;
        }
    }
    printf ("Internal Error at %s:%d - code: %s\n", func, fn_e, code_s);
}

// Generic add to the front of a list
void AddFileDatum (FILEDATA **List, char *hfile, int *lineno, char *HRefPath)
{
    const char *func = "AddFileDatum";
    FILEDATA *New;

    New = malloc (sizeof (*New));
    if (!New) {
        InternalError (func, 1, IERR_MALLOC);
        goto done;
    }

    New->next = *List;
    strcpy (New->ht_file, hfile);
    New->ht_lineno = *lineno;
    strcpy (New->ref_file, HRefPath);

    *List = New;
done: ;
}

